"""Utilities."""
import inspect
import itertools
import re
from functools import partial
from typing import Any, Callable, List, Optional, Set, Union, get_type_hints

import falcon

from falcontyping.typedjson import args_of, origin_of

from .exceptions import TypeValidationError

try:
    from typing_extensions import Protocol

    class ResourceMethodWithoutReturnValue(Protocol):

        def __call__(self, request: falcon.Request, respone: falcon.Response, **kwargs: Any) -> None:
            ...

    class ResourceMethodWithReturnValue(Protocol):

        def __call__(self, request: falcon.Request, respone: falcon.Response, **kwargs: Any) -> Any:
            ...

except ImportError:
    ResourceMethodWithoutReturnValue = Callable  # type: ignore
    ResourceMethodWithReturnValue = Callable     # type: ignore

try:
    from pydantic import BaseModel as PydanticBaseModel

except ImportError:
    PydanticBaseModel = None  # type: ignore


try:
    from marshmallow import Schema as MarshmallowSchema

except ImportError:
    MarshmallowSchema = None  # type: ignore


_AVAILABLE_EXTERNAL_SCHEMA_PRODIVERS = set(
    filter(None, [MarshmallowSchema, PydanticBaseModel])
)

# Taken from https://github.com/falconry/falcon/blob/master/falcon/routing/compiled.py#L27
_FIELD_PATTERN = re.compile(
    # NOTE(kgriffs): This disallows the use of the '}' character within
    # an argstr. However, we don't really have a way of escaping
    # curly brackets in URI templates at the moment, so users should
    # see this as a similar restriction and so somewhat unsurprising.
    #
    # We may want to create a contextual parser at some point to
    # work around this problem.
    r'{((?P<fname>[^}:]*)((?P<cname_sep>:(?P<cname>[^}\(]*))(\((?P<argstr>[^}]*)\))?)?)}'
)

_METHOD_NAMES: List[str] = [
    'on_%s' % method for method in ('get', 'post', 'put', 'patch', 'delete', 'head', 'options')
]


def validate_type_preconditions(type_: Any) -> None:
    """Validate that used type hints are supported."""

    error = ('Resource methods must accept/return either Nothing, '
             'marshmallow.Schema or pydantic.BaseModel not {}')

    def predicate(type_: Any) -> bool:
        """Check if type can be used in the body of the request or response."""
        origin, args = origin_of(type_), args_of(type_)

        if origin is Union:
            return all(predicate(arg) for arg in args)

        else:
            has_supported_parent = any(
                issubclass(type_, parent) for parent in _AVAILABLE_EXTERNAL_SCHEMA_PRODIVERS
            )
            return (has_supported_parent or type_ == type(None)) and type_ is not Any  # noqa

    try:
        if not predicate(type_):
            raise TypeValidationError(error.format(type_))

    except TypeError:
        raise TypeValidationError(error.format(type_))


def validate_method_signature(method: ResourceMethodWithReturnValue, uri_parameters: Set[str]) -> Optional[str]:
    """Validate whether resource method has the right signature."""
    hints = get_type_hints(method)

    for parameter in filter(lambda p: hints.get(p, Any) is Any, uri_parameters):
        raise TypeValidationError(f'URI parameter {parameter} has no type hint or is not part of method signature')

    validate_type_preconditions(
        hints.pop('return')
    )

    if len(inspect.getfullargspec(method).args) < 3:
        raise TypeValidationError('Every resource method must have the first two parameters as '
                                  'falcon.Request and falcon.Response')

    body_parameters = set(hints) - (uri_parameters | set(itertools.islice(hints.keys(), 2)))

    if len(body_parameters) > 1:
        raise TypeValidationError('Any resource method can not accept more than one '
                                  'marshmallow.Schema or pydantic.BaseModel as a body parameter')

    for parameter_name in body_parameters:
        validate_type_preconditions(hints[parameter_name])

    return next(iter(body_parameters), None)


def patch_resource_methods(uri_template: str, resource: Any) -> None:
    """Patch resource methods to add type hint supports."""
    uri_parameters = set(match.group('fname') for match in _FIELD_PATTERN.finditer(uri_template))

    def resource_method_wrapper(method: ResourceMethodWithReturnValue) -> ResourceMethodWithoutReturnValue:
        """
        Wraps resource methods.

        This function will takes whatever is returned by resource methods and assigns
        it to response.media.
        """
        def curried(request: falcon.Request, response: falcon.Response, **kwargs: Any) -> None:
            response.media = method(request, response, **kwargs)

        return curried  # type: ignore

    resource.methods_body_parameters = {}
    for method_name in filter(partial(hasattr, resource), _METHOD_NAMES):

        method: ResourceMethodWithReturnValue = getattr(resource, method_name)

        if not callable(method):
            raise TypeValidationError(f'{resource}.{method_name} must be a Callable')

        try:
            resource.methods_body_parameters[method_name] = validate_method_signature(method,
                                                                                      uri_parameters=uri_parameters)

        except TypeValidationError as e:
            raise TypeValidationError(f'{resource}.{method} raised: {e}') from None

        wrapped = resource_method_wrapper(method)

        # Preserve documentation and annotations
        wrapped.__name__ = method.__name__  # type: ignore
        wrapped.__doc__ = method.__doc__
        wrapped.__annotations__ = method.__annotations__

        setattr(resource, method_name, wrapped)



"""Typing middleware."""
from typing import Any, Callable, Dict, Optional, get_type_hints

import falcon
from falcon import Request, Response

from falcontyping.base import (MarshmallowSchema, PydanticBaseModel,
                               TypedResource, TypeValidationError,
                               schema_providers)
from falcontyping.typedjson import DecodingError, ExternalSerializerException
from falcontyping.typedjson import decode as decode_using_hints

_VALID_RESPONSE_TYPES = schema_providers | set([dict])


class TypingMiddleware:

    @staticmethod
    def _decode_or_raise_error(hint: Any, parameter: Any) -> Any:
        """
        Decode value using type hint or fail.

        :raises: falcon.HTTPError
        """
        result = decode_using_hints(hint, parameter)

        if isinstance(result, DecodingError):
            if isinstance(result.reason, ExternalSerializerException):
                raise result.reason.exception from None

            else:
                raise falcon.HTTPError(status=falcon.HTTP_UNPROCESSABLE_ENTITY,  # pylint: disable=no-member
                                       description=f'\'{parameter}\' must be of type {hint} not {type(parameter)}')

        return result

    def process_request(self, request: Request, response: Response) -> None:
        """
        Process the request before routing it.

        Because Falcon routes each request based on req.path, a
        request can be effectively re-routed by setting that
        attribute to a new value from within process_request().

        :param request: Request object that will eventually be
            routed to an on_* responder method.
        :param response: Response object that will be routed to
            the on_* responder.
        """
        ...

    def process_resource(self, request: Request, response: Response, resource: Any, parameters: Dict) -> None:
        """
        Process the request after routing.

        This method is only called when the request matches
        a route to a resource.

        :param request: Request object that will be passed to the
            routed responder.
        :param response: Response object that will be passed to the
            responder.
        :param resource: Resource object to which the request was
            routed.
        :param parameters: A dict-like object representing any additional
            parameters derived from the route's URI template fields,
            that will be passed to the resource's responder
            method as keyword arguments.
        """
        if not isinstance(resource, TypedResource):
            return

        handler: Optional[Callable] = getattr(resource, 'on_%s' % request.method.lower(), None)

        if handler:
            # Get hints for only those variables that should be passed to the request handler.
            hints = get_type_hints(handler)

            # Decode values using type hints, All values in parameters will be based as
            # Keyword arguments to the request handler.
            for parameter in parameters:
                parameters[parameter] = self._decode_or_raise_error(hints[parameter], parameters.get(parameter))

            # Decode body parameter if there is one.
            body_parameter = resource.methods_body_parameters[handler.__name__]
            if body_parameter:
                parameters[body_parameter] = self._decode_or_raise_error(hints[body_parameter],
                                                                         getattr(request, 'media', None))

    def process_response(self, request: Request, response: Response, resource: Any, request_succeeded: bool) -> None:
        """
        Post-processing of the response (after routing).

        :param request: Request object.
        :param response: Response object.
        :param resource: Resource object to which the request was routed.
        May be None if no route was found for the request.

        :param request_succeeded: True if no exceptions were raised while the framework processed and
        routed the request; otherwise False.
        """
        if not (isinstance(resource, TypedResource) and request_succeeded):
            return

        handler: Optional[Callable] = getattr(resource, 'on_%s' % request.method.lower(), None)

        if handler:
            # Get type hint for the return type of the request handler.
            hint: Any = get_type_hints(handler)['return']

            # Decode returned value using the "return" type hint.
            media = getattr(response, 'media', None)
            media = decode_using_hints(hint, media)

            if isinstance(media, DecodingError):
                raise TypeValidationError(
                    f'{resource}.{handler} returned a value of type {media}:{type(media)} instead of {hint}')

            if media is None:
                ...

            else:
                if not any(isinstance(media, type_) for type_ in _VALID_RESPONSE_TYPES):
                    raise TypeValidationError(f'{resource}.{handler} returned a unexpected value. ',
                                              f'Resource methods must return either Nothing, '
                                              f'marshmallow.Schema or pydantic.BaseModel not {type(media)}')

                try:
                    if isinstance(media, dict):
                        ...

                    elif isinstance(media, PydanticBaseModel):
                        media = media.dict()

                    elif isinstance(media, MarshmallowSchema):
                        media = hint().dump(media)

                    response.media = media

                except Exception:
                    raise TypeValidationError(f'{resource}.{handler} '
                                              f'failed to serialize {media}:{type(media)} to JSON.')



"""
The MIT License (MIT)

Copyright (c) 2018 Tomoya Kose

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

This is a modified version of the decoding module at:
https://github.com/mitsuse/typedjson-python/blob/master/typedjson/decoding.py

Author: Tomoya Kose <tomoya@mitsuse.jp>
Repository: https://github.com/mitsuse/typedjson-python
"""
from typing import Any, Tuple, TypeVar, Union

Decoded = TypeVar('Decoded')
Value = TypeVar('Value')

Path = Tuple[str, ...]


class TypeMismatch(Exception):
    def __init__(self, path: Path) -> None:
        self.__path = path

    def __eq__(self, x: Any) -> bool:
        if isinstance(x, TypeMismatch):
            return self.path == x.path

        else:
            return False

    def __str__(self) -> str:
        return f'<TypeMismatch path={self.path}>'

    @property
    def path(self) -> Path:
        return self.__path


class UnsupportedDecoding(Exception):
    def __init__(self, path: Path) -> None:
        self.__path = path

    def __eq__(self, x: Any) -> bool:
        if isinstance(x, UnsupportedDecoding):
            return self.path == x.path

        else:
            return False

    def __str__(self) -> str:
        return f'<UnsupportedDecoding path={self.path}>'

    @property
    def path(self) -> Path:
        return self.__path


class ExternalSerializerException(Exception):
    def __init__(self, path: Path, exception: Exception) -> None:
        self.__path = path
        self.exception = exception

    def __eq__(self, x: Any) -> bool:
        if isinstance(x, ExternalSerializerException):
            return self.path == x.path
        else:
            return False

    def __str__(self) -> str:
        return f"<ExternalSerializerException path={self.path}>"

    @property
    def path(self) -> Path:
        return self.__path


FailureReason = Union[TypeMismatch, UnsupportedDecoding, ExternalSerializerException]


class DecodingError(Exception):
    def __init__(self, reason: FailureReason) -> None:
        self.__reason = reason

    def __eq__(self, x: Any) -> bool:
        if isinstance(x, DecodingError):
            return self.reason == x.reason
        else:
            return False

    def __str__(self) -> str:
        return f"<DecodingError reason={self.reason}>"

    @property
    def reason(self) -> FailureReason:
        return self.__reason
